// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: fwd_table.h,v 1.12 2002/09/30 04:50:25 carzanig Exp $
//
#ifndef FWD_TABLE_H
#define FWD_TABLE_H

#include "fwdconf.h"
#include "siena_types.h"
#include "tst.h"
#include "ft_allocator.h"
#include "bitvector.h"

using namespace std;

/** \mainpage Siena Fast Forwarding Documentation
 *
 *  This documentation describes the implementation of the Fast
 *  Forwarding module of <a
 *  href="http://www.cs.colorado.edu/serl/siena/">Siena</a>.  This
 *  implementation consists of a fowarding table and a driver for it.
 *
 *  <p>A forwarding table is an association of <em>predicates</em> to
 *  <em>interfaces</em>, one predicate is associated with one
 *  interface.  A predicate is a <em>disjunction</em> of
 *  <em>conjunctions</em> of elementary <em>constraints</em>.  The
 *  forwarding table provides essentially one match function that
 *  takes a <em>message</em> <em>M</em> and outputs the set of
 *  interfaces associated with predicates matched by <em>M</em>.
 *
 *  <p>The driver is a function that reads commands from a file.  The
 *  commands understood by the driver include essentially all the
 *  operations of the forwarding table.  In particular, the driver can
 *  add associations of predicates to interfaces, and can request the
 *  processing of messages.  Examples of commands accepted by the
 *  driver are available from the <a href="examples.html">Example
 *  Section</a>.
 **/

/** \example "input commands"
 *
 *  This is an example of commands understood by the driver of the
 *  forwarding table:
 * 
 *  <pre>
 *  %timer_start
 *  ifconfig 1 name = "Antonio" | name = "Alex", nationality = "italian"
 *  ifconfig 2 airline = "UA", price < 500 | dest = "MXP", price < 700
 *  %timer_stop
 *  %print_statistics
 *  %timer_start
 *  select name="Antonio" nationality="italian"
 *  select name="Alex"
 *  select airline="UA" price=200 dest="IAD"
 *  %timer_stop
 *  %print_statistics
 *  </pre>
 **/

/** Interface type.  
 *
 *  For simplicity, an interface is represented by a number.  This is
 *  an explicit design choice.  More information can be associated
 *  with an interface through a secondary data structure (a vector or
 *  anything else).
 **/
typedef unsigned int		if_t;

/** inteface identifier within the matching algorithm.  
 *
 *  As opposed to <code>if_t</code> which identifies user-specified
 *  interface numbers, this is going to be used for the identification
 *  of interfaces within the matching algorithm, which may require a
 *  set of contiguous identifiers.  So, for example, the user may
 *  specify interfaces 6, 78, and 200, while the internal
 *  identification would be 0, 1, and 2 respectively (or similar).  I
 *  treat it as a different type (from <code>if_t</code>) because I
 *  don't want to mix them up (now that I write this, I'm not even
 *  sure the compiler would complain. Oh well.)
 **/
typedef unsigned int		ifid_t;

/** the matching function doesn't actually produce an output.
 *  Instead, it delegates the processing of matching interfaces to a
 *  specialized "output" processor.  This base class defines the
 *  interface if such an output processor.
 **/
class IProcessor {
public:
    /** output function.
     *
     *  This function is called within the matching function of the
     *  forwarding table.  The parameter is the interface identifier
     *  of the matching interface.  This function may explicitly cause
     *  the matching function to terminate by returning true result.
     **/
    virtual bool		output(if_t) = 0;
};

class Constraint;
class Selectivity;
class Interface;
class Attribute;
class Filter;

/** The heart of the forwarding table. 
 *
 *  A <code>FwdTable</code> associates predicates to interfaces.  A
 *  Predicate is a disjunction of conjunctions of elementary
 *  constraints.  An interface is simply an identifier, corresponding
 *  to the link-level address of a neighbor.
 *
 *  <p>The <code>FwdTable</code> is a <em>dictionary</em> data
 *  structure, meaning that the set of associations is constructed by
 *  adding associations, but can not be modified by removing
 *  individual associations.  
 *
 *  <p>Once constructed, the forwarding table can match incoming
 *  messages.  The forwarding table delegates the processing of
 *  matched notifications to a given <em>output processor</em> object.
 *
 *  <p>The forwarding table manages its own memory allocation through
 *  a block-allocator.  The forwarding table can be completely cleared
 *  and built over again.  When cleared, the forwarding table releases
 *  all the previously allocated memory.
 **/
class FwdTable {
public:
    /** constructs a forwarding table with a private memory allocator
     *	that uses memory chunks of the given size.
     **/
				FwdTable(unsigned int memblk_size = 65536);

    /** associates a predicate to an interface.
     *
     *  This is the method that constructs the forwarding table.  This
     *  method must be called <em>once</em> for each interface, after
     *  the forwarding table is constructed or after it has been
     *  cleared.  Using this method twice on the same interface
     *  without clearing the forwarding table has undefined effects.
     **/
    void			ifconfig(if_t, const sx_predicate &);

    /** processes a message, calling the output() function on the
     *	given <code>IProcessor</code> object for each matching interface.
     **/
    void			match(const sx_message &, IProcessor &) const;

    /** clears the forwarding table.
     *
     *  This method removes all the associations from the forwarding
     *  table and releases allocated memory.
     **/
    void			clear();

    /** footprint of the forwarding table.
     *
     *  returns the number of bytes of memory allocated by the
     *  forwarding table.
     **/
    unsigned int		bytesize() const;

    /** determines the number of pre-processing rounds applied to
     *	every message.
     **/
    void			set_preprocess_rounds(unsigned int);

    /** returns the current number of pre-processing rounds.
     **/
    unsigned int		get_preprocess_rounds() const;

private:
    /** private allocator of the forwarding table.  
     *
     *  All the data structure forming the forwarding table are
     *  allocated through this memory management system.
     **/
    FTAllocator			memory;

    /** total number of interfaces associated with a predicate in the
     *  forwarding table.
     **/
    ifid_t			ifcount;

    /** number of pre-processing rounds
     **/
    unsigned int		preprocess_rounds;

    /** main index of constraints
     **/
    TST				attributes;

    /** first element in the <em>selectivity table</em>
     *
     *  this is the element with the highest selectivity level.  In
     *  other words, this is the constraint name that, if not present
     *  in the message, can exclude the higher number of interfaces.
     **/
    Selectivity *		selectivity_first;

    /** last element in the <em>selectivity table</em>
     *
     *  this is the element with the lowest selectivity level.
     **/
    Selectivity *		selectivity_last;

    /** pre-processing function
     **/
    void			preprocess_message(const sx_message &, 
						   bitvector *, 
						   unsigned int) const;

    /** adds a new item to the selectivity table.
     *
     *  The selectivity table is a double-link list of
     *  <code>Selectivity</code> objects.  The list is sorted according to
     *  the level of selectivity, which correspond to the size of the
     *  <code>exclude</code> set.
     *
     *  <p>When a new selectivity item is created, it is added at the
     *  bottom of the list, having selectivity level 1.
     **/
    Selectivity	*		new_selectivity(const string &, ifid_t);

    /** adds an interface to the set of excluded interfaces of an
     *	existing item in the selectivity table 
     *
     *  When a new interface is added to the <code>exclude</code> set
     *  of a selectivity item of a given name, that item is moved
     *  along the selectivity list in order to maintain the list
     *  sorted by level of selectivity.  This is done by shifting the
     *  item as in a bubble-sort algorithm.
     **/
    void			add_to_selectivity(Selectivity *, ifid_t);

    /** utility function that finds an attribute in the constraint index **/
    Attribute *			get_attribute(const sx_name & name);

    /** utility function that connects a constraint descriptor to the
     *  descriptor of the filter containing that constraint
     **/
    void			connect(Constraint *, Filter *);
};

/** this constraint processor represents the main "iterator" of the
 *  matching function
 **/
class CProcessor {
public:
    /** processes one matched constraint.  Resturns true when the
     *  matching process can terminate.
     **/
    virtual bool		process_constraint(const Constraint *) = 0;
};

/** filter descriptor
 *
 *  Represents a conjunction of constraints.  
 **/
class Filter {
public:
    /** constructs a filter descriptor associating the filter with the
     *  given interface 
     **/
    Filter(Interface * i);

    /** descriptor of the interface associated with this filter 
     *
     *  The current implementation is not capable of recognizing two
     *  identical filters.  Therefore, every filter will generate a
     *  filter descriptor like this, and therefore a filter descriptor
     *  is associated with a single interface.  This is why this is
     *  simply a pointer to an interface descriptor.
     *
     *  <p>At some point, we migh want to be able to figure out that
     *  F1==F2, and therefore to allow each filter to refer to more
     *  than one interfaces.  In that case, we will have to maintain a
     *  set of pointers to interface descriptors.
     **/
    Interface *			i;

    /** number of constraints composing this filter
     *
     *  Notice that the implicit restriction of this declaration is
     *  that we do not allow more than 255 constraints per filter.
     **/
    unsigned char		size;
};

/** link in a single-link list for filters **/
class flist {
public:
    /** the filter **/
    const Filter * f;
    /** link to the next filter in the list **/
    const flist * next;

    /** constructor **/
    flist(Filter * xf, flist * xn);
};

/** constraint descriptor
 *
 *  it is essentially a set of pointers to the filters in which this
 *  constraint appears.
 **/
class Constraint {
public:
    /** list of filters associated with this constraint
     *
     *  All the filters in which this constraint appears.
     *
     *  <p>Right now, we add filters to this list as they are passed
     *  to the forwarding table.  It might be possible and
     *  advantageous to figure out an optimal ordering of this list in
     *  order to speed up the maching process.  For example, we could
     *  sort the filters based on their size.
     **/
    flist *			filters;
    /** constructor **/
    Constraint();
};

#ifdef HAVE_INLINE
#include "fwd_table.icc"
#endif

#endif
